package com.brilliance.mda.support.mybatis;

import com.brilliance.mda.runtime.mda.*;
import com.brilliance.mda.runtime.mda.driver.MdaDriver;
import com.brilliance.mda.runtime.mda.driver.MdaEnv;
import com.brilliance.mda.runtime.mda.impl.AbstractDaoSession;
import com.brilliance.mda.runtime.mda.impl.Argument;
import com.brilliance.mda.support.mybatis.count.CounterService;
import com.brilliance.mda.support.mybatis.dync.mapper.DbExecuteMapper;
import org.mybatis.spring.SqlSessionTemplate;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;

import java.io.Serializable;
import java.sql.SQLException;
import java.util.*;

import static com.brilliance.mda.support.mybatis.MybatisArgumentAdapter.AdaType.DELETE;
import static com.brilliance.mda.support.mybatis.MybatisArgumentAdapter.AdaType.SELECT;
import static com.brilliance.mda.support.mybatis.MybatisArgumentAdapter.HOLDER_SQL;

/**
 * 基于Spring SqlSessionTemplate的实现.需要满足以下要求：
 * 1，src/main/resources/mapper中的xml文件需要加入到mapper文件的扫描路径中
 * 2，所有mybatis的mapper.xml文件namespace命名规则遵循
 * "[namespacePackage].[BO.class.getSimpleName()]Mapper"
 * namespacePackage : 类属性
 * BO.class.getSimpleName(): BO类的名称
 * <p>
 * 3，所有的mybatis的mapper.xml文件中包含以下两类sql模板
 * a，基于Mybatis官方Generator生成的： insert,updateByPrimaryKey,delete
 * b，特殊的:
 * deleteByIds（采用in方式删除主键匹配）
 * <delete id="deleteByIds" parameterClass="java.util.List">
 * sql...
 * </delete>
 * <p>
 * dyncRead(实现Argument方式的动态查询)
 * <select id="dyncRead" >
 * ${sql}
 * </select>
 * <p>
 * dyncDelete(实现Argument方式的动态删除)
 * <delete id="dyncDelete" >
 * ${sql}
 * </delete>
 *
 * @author hulei
 */
@Component
public class MyBatisDaoSession extends AbstractDaoSession implements IDaoSession {

    private static Logger log = LoggerFactory.getLogger(MyBatisDaoSession.class);
    private static ThreadLocal<ResultSet> resultLocal = new ThreadLocal<>();

    //定义mapper文件中的sql语句id
    private static final String SQLID_INSERT = "insert";
    private static final String SQLID_UPDATE = "updateByPrimaryKey";
    private static final String SQLID_DELETE = "deleteByPrimaryKey";
    private static final String SQLID_DELBYIDS = "deleteByIds";
    private static final String SQLID_DYNCREAD = "dyncRead";
    private static final String SQLID_DYNCDEL = "dyncDelete";


    @Autowired
    CounterService counter;

    @Autowired
    DbExecuteMapper dbExecuteMapper;

    @Autowired
    SqlSessionTemplate sqlSessionTemplate;

    /**
     * mapper.xml文件中namespace指向类的package
     */
    private String namespacePackage = "com.ceb.gjjs.mapper";

    public void setNamespacePackage(String namespacePackage) {
        this.namespacePackage = namespacePackage;
    }

    /**
     * mybatis sqlid 格式化模板，
     * %s: namespackPackage
     * %s: BO类名
     * %s：具体sqlId
     */
    private String sqlIdTemplate = "%s.%sMapper.%s";

    public void setSqlIdTemplate(String sqlIdTemplate) {
        this.sqlIdTemplate = sqlIdTemplate;
    }

    @Override
    public int dbCounter(String seqname) {
        return this.counter.dbCounter(seqname);
    }

    @Override
    public int dbInsert(IModule module) {
        String sqlId = getSqlId(module.getClass(), SQLID_INSERT);
        return executeUpdate(sqlId, module);
    }

    @Override
    public int dbUpdate(IModule module, Object v1, String whereCol1) {
        return 0;
    }

    @Override
    public int dbUpdate(IModule module, Object v1, String whereCol1, Object v2, String whereCol2) {
        return 0;
    }

    @Override
    public <T extends IModule> int dbUpdate(T module) {
        String sqlId = getSqlId(module.getClass(), SQLID_UPDATE);
        return executeUpdate(sqlId, module);
    }

    @Override
    public int dbDelete(IModule module) {
        String sqlId = getSqlId(module.getClass(), SQLID_DELETE);
        return executeUpdate(sqlId, module);
    }

    @Override
    public int dbDelete(IModule m, Argument<? extends Object>... args) {
        Class clazz = (Class) m.getClass();
        MybatisArgumentAdapter adapter = new MybatisArgumentAdapter(DELETE, clazz, args);
        return executeUpdate(getSqlId(clazz, SQLID_DYNCDEL), adapter.getSqlParams());
    }

    public <T extends IModule> int dbDelete(T module, Serializable... id) {
        Class<T> clazz = (Class<T>) module.getClass();
        String sqlId = getSqlId(clazz, SQLID_DELBYIDS);
        return executeUpdate(sqlId, Arrays.asList(id));
    }

    public <T extends IModule> int dbDelete(Class<T> clazz, Map paramMap) {
        throw new UnsupportedOperationException("暂不支持该方法！");
    }

    @Override
    public int dbDelete(IModule module, Object val, String col) {
        return 0;
    }

    @Override
    public int dbDelete(IModule module, Object val, String col, Object val2, String col2) {
        return 0;
    }

    @Override
    public <T extends IModule> T dbRead(T module, Argument... args) {
        MybatisArgumentAdapter adapter = new MybatisArgumentAdapter(SELECT, module.getClass(), args);

        T entity = this.dyncReadOne(module.getClass(), adapter);
        if (entity != null) {
            MdaDriver.copyValues(module, entity);
        }
        return entity;
    }

    @Override
    public <T extends IModule> T dbRead(T module, Object... args) {
        return null;
    }

    @Override
    public <T extends IModule> T dbReadNoData(T module, Object... args) {
        return null;
    }

    @Override
    public <T extends IModule> T dbReadByArguments(T module, Argument<?>... args) {
        return null;
    }

    @Override
    public <T extends IModule> T dbReadHold(T module, Object... args) {
        return null;
    }

    @Override
    public <T extends IModule> int dbReadSet(IModuleList<T> list, Object... args) {
        return 0;
    }

    @Override
    public <T extends IModule> int dbReadset(IModuleList<T> list, int limitSize, Argument... args) {
        list.clear();
        MybatisArgumentAdapter adapter = new MybatisArgumentAdapter(SELECT, list.getDataClass(), args);
        String temp = adapter.getSqlTemplate();
        Map<String, Object> params = adapter.getSqlParams();

        if (limitSize == 0) {
            List<T> resultList = this.dyncRead(list.getDataClass(), params);
            list.addAll(resultList);
        } else {
            //TODO 实现分页查询
//			Page<T> page = new Page<>(1,limitSize);
//			IPage<T> pageData = baseMapper.selectPage(page, wrapper);
//			list.addAll(pageData.getRecords());
        }
        return list.size();
    }

    @Override
    public <T extends IModule> int dbReadset(IModuleList<T> list, Object... args) {
        list.clear();
        if (args instanceof Argument[]) {
            this.dbReadset(list, 0, (Argument[]) args);
        }
        return list.size();
    }

    @Override
    public <T extends IModule> int dbReadset(IModuleList<T> list, String whereSql, Object[] objects) {

        list.clear();
        MybatisArgumentAdapter adapter = new MybatisArgumentAdapter(SELECT, list.getDataClass(), whereSql, objects);
        List<T> result = this.dyncRead(list.getDataClass(), adapter);
        list.addAll(result);
        return list.size();

    }

    @Override
    public <T extends IModule> int dbReadset(IModuleList<T>[] lists, int maxSize, String whereClause, Object[] datas) {
        List<Class<? extends IModule>> clazzList = new ArrayList<>();
        for (IModuleList iModule : lists) {
            clazzList.add(iModule.getDataClass());
        }
        MybatisArgumentAdapter adapter = new MybatisArgumentAdapter(SELECT, clazzList, whereClause, datas);
        for (IModuleList iModule : lists) {
            List<T> result = this.dyncRead(iModule.getDataClass(), adapter);
            iModule.addAll(result);
        }
        return lists[0].size();
    }

    @Override
    public void dbReadset(IModuleList[] lists, String whereClause, Object[] datas) {

    }

    @Override
    public void dbReadset(IModuleList list, int maxSize, String sql) {

    }

    /**
     * 方法存在SQL注入风险，不建议使用
     */
    @Deprecated
    @Override
    public <T extends IModule> int dbReadset(IModuleList<T> list, String whereSql) {
        return this.dbReadset(list, whereSql, null);
    }


    @Override
    public <T extends IModule> int dbReadSetByArguments(IModuleList<T> list, Argument<?>... args) {
        return 0;
    }

    @Override
    public void dbConnect() {
        sqlSessionTemplate.getSqlSessionFactory().openSession(false);
    }


    @Override
    public void dbDisconnect() {
        try {
            sqlSessionTemplate.getConnection().close();
        } catch (SQLException e) {
            e.printStackTrace();
        }
    }


    @Override
    public void dbExecuteSQL(IStream stm) {

    }

    @Override
    public void dbExecuteSQL(IStream sql, IStream retstream) {

    }

    @Override
    public String dbName() {
        return sqlSessionTemplate.getConfiguration().getDatabaseId();
    }


    @Override
    public void dbExecuteSQL(String sql, Object... objects) {
        Map<String, Object> map = new HashMap<>();
        if (objects != null) {
            sql = MybatisArgumentAdapter.convertPlaceHolder(sql, objects, map);
        }
        map.put(HOLDER_SQL, sql);
        if (sql.toUpperCase().startsWith("SELECT")) {
            List<Map<String, Object>> result = dbExecuteMapper.dyncReadForMap(map);
            if (result == null || result.size() == 0) {
                setNoMoreRow();
                return;
            }
            resultLocal.set(new ResultSet(result));
        } else {
            int cnt = dbExecuteMapper.dyncUpdateForMap(map);
        }
    }

    @Override
    public void dbFetch(IModule module) {

    }

    @Override
    public void dbFetchStream(IStream args) {

    }


    @Override
    public void dbFetchFields(Object... arguments) {
        if (arguments == null || arguments.length == 0) return;
        if (resultLocal.get() == null) {
            log.error("must execute dbExecuteSQL first !");
            setNoMoreRow();
            return;
        }
        if (!resultLocal.get().hasNext()) {
            setNoMoreRow();
            return;
        }
        Map<String, Object> row = resultLocal.get().next();

        for (Object o : arguments) {
            if (o instanceof Argument) {
                Argument argument = (Argument) o;
                setArgumentValue(argument, row.get(argument.getFieldName().toUpperCase()));
            }
        }
    }

    @Override
    public void dbFetchFields(Argument<?>... arguments) {

    }


    private void setArgumentValue(Argument<? extends Object> argument, Object obj) {
        if (obj == null) return;
        Argument<Object> arg = (Argument<Object>) argument;
        arg.value = obj;
    }

    @Override
    public void dbCloseCursor() {
        //mybatis方式不直接访问结果集，不需要关闭结果集
    }

    @Override
    public String dbSqlDate(Date endDat) {
        return null;
    }

    @Override
    public void dbFreeAll() {

    }

    @Override
    public void setWaitCursor() {

    }

    @Override
    public void dbHold(IModule module) {

    }

    @Override
    public void dbFree(IModule module) {

    }

    @Override
    public void dbSelect(IModule module, String string, String sql, Object... val) {

    }

    @Override
    public void DBOpen(IModule module) {

    }

    @Override
    public void dbSelectCursor(IModule module, String whereSql, Object... params) {

    }

    @Override
    public void putDao(Class<?> key, Object dao) {
        throw new UnsupportedOperationException("不支持的方法！");
    }


    private <T> String getSqlId(Class<T> clazz, String sid) {
        return String.format(sqlIdTemplate, namespacePackage, clazz.getSimpleName(), sid);
    }


    private <T> T dyncReadOne(Class<? extends IModule> clazz, MybatisArgumentAdapter adapter) {

        String sqlId = getSqlId(clazz, SQLID_DYNCREAD);
        T entity = sqlSessionTemplate.selectOne(sqlId, adapter.getSqlParams());
        if (entity == null) {
            setNoMoreRow();
        }
        return entity;
    }

    private <T> List<T> dyncRead(Class<T> clazz, MybatisArgumentAdapter adapter) {

        String sqlId = getSqlId(clazz, SQLID_DYNCREAD);
        List<T> list = sqlSessionTemplate.selectList(sqlId, adapter.getSqlParams());
        if (list == null || list.size() == 0) {
            setNoMoreRow();
        }
        return new ArrayList<>(list);
    }


    private <T> List<T> dyncRead(Class<T> clazz, Map params) {

        String sqlId = getSqlId(clazz, SQLID_DYNCREAD);
        List<T> list = sqlSessionTemplate.selectList(sqlId, params);
        if (list == null || list.size() == 0) {
            setNoMoreRow();
        }
        return new ArrayList<>(list);
    }

    public static class ResultSet {
        private int index;
        private List<Map<String, Object>> data;

        public ResultSet(List<Map<String, Object>> data) {
            this.index = 0;
            this.data = data;
        }

        public Map<String, Object> next() {
            if (index < data.size()) {
                return data.get(index++);
            }
            return null;
        }

        public boolean hasNext() {
            return data != null && index < data.size();
        }
    }

    private void setNoMoreRow() {
        MdaEnv.getContext().setErrorCode(Constants.NO_MORE_ROW);
    }

    private int executeUpdate(String mybatisSqlId, Object params) {
        int cnt = sqlSessionTemplate.update(mybatisSqlId, params);
        return cnt;
    }

}
